from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

import numpy as np

PRECISION = 12
"""The default precision to use for floats in configurations."""


@dataclass
class Config(Mapping[str, Any]):
    """A configuration to evaluate."""

    config_id: str
    """Some unique identifier"""

    values: dict[str, Any] | None
    """The actual config values to evaluate.

    In the case this config was deserialized, it will likely be `None`.
    """

    description: str | None = None
    """A description of the configuration."""


    def to_dict(self) -> dict[str, Any]:
        """Convert the configuration values to a dictionary.

        Returns:
            A dictionary of the configuration values.
        """
        assert self.values is not None
        return self.values


    def to_tuple(self, precision: int | None = None) -> tuple:
        """Convert the configuration values to a tuple with specified precision.

        Args:
            precision: The precision to round the float values to.
                If `None`, the default
                [`PRECISION`][hpoglue.config.PRECISION] is used.

        Returns:
            A tuple of the configuration values with the specified precision.
        """
        if precision is None:
            precision = PRECISION

        assert self.values is not None
        return tuple(self.set_precision(self.values, precision).values())

    @staticmethod
    def set_precision(values: dict, precision: int) -> dict[str, Any]:
        """Set the precision of float values in the configuration for continuations.

        Args:
            values: The dictionary of configuration values.
            precision: The precision to round the float values to.

        Returns:
            The dictionary with float values rounded to the specified precision.
        """
        # NOTE: Make sure not to edit the dictionary in place as we return a value.
        return {k: np.round(v, precision) if isinstance(v, float) else v for k, v in values.items()}

    def __getitem__(self, key: str) -> Any:
        assert self.values is not None
        return self.values[key]

    def __iter__(self):
        assert self.values is not None
        return iter(self.values)

    def __len__(self) -> int:
        assert self.values is not None
        return len(self.values)


from hpoglue.benchmark import (
    Benchmark,
    BenchmarkDescription,
    FunctionalBenchmark,
    SurrogateBenchmark,
    TabularBenchmark,
)
from hpoglue.config import Config
from hpoglue.measure import Measure
from hpoglue.optimizer import Optimizer
from hpoglue.problem import Problem
from hpoglue.query import Query
from hpoglue.result import Result
from hpoglue.run_glue import run_glue as run

__all__ = [
    "Benchmark",
    "BenchmarkDescription",
    "Config",
    "FunctionalBenchmark",
    "Measure",
    "Optimizer",
    "Problem",
    "Query",
    "Result",
    "SurrogateBenchmark",
    "TabularBenchmark",
    "run",
]



from __future__ import annotations

import logging
import subprocess
import warnings
from collections.abc import Sequence
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any

from hpoglue.utils import env_pkg_version_compat

logger = logging.getLogger(__name__)

WHEN_MODULE_INITIALIZED = datetime.now().isoformat()

GLUE_REPO = "github.com/automl/hpoglue.git"
GLUE_PYPI = "hpoglue"
GLUE_GIT_SSH_INSTALL = "git+ssh://git@github.com/automl/hpoglue.git"


@dataclass
class Env:
    """Represents an hpoglue environment configuration with a name, Python version,
    requirements, and post-install commands.
    """
    name: str
    """The name of the environment."""

    python_version: str = field(default="3.10", repr=False)
    """The Python version of the environment."""

    requirements: tuple[str, ...] = field(default=(), repr=False)
    """Environment requirements."""

    post_install: tuple[str, ...] = field(default=(), repr=False)
    """Post-install commands."""

    def __post_init__(self) -> None:
        match self.requirements:
            case tuple():
                pass
            case str():
                self.requirements = (self.requirements,)
            case None:
                self.requirements = ()
            case _:
                raise ValueError(
                    f"Invalid requirements type: {type(self.requirements)}, expected tuple!"
                )
        match self.post_install:
            case tuple():
                _cmd = []
                for command in self.post_install:
                    command = command.replace("python", "").replace("-m", "").strip()  # noqa: PLW2901
                    if command:
                        _cmd.append(command)
                self.post_install = tuple(_cmd)
            case str():
                self.post_install = (
                    self.post_install.replace("python", "")
                    .replace("-m", "")
                    .strip()
                )
                self.post_install = (self.post_install,)
            case None:
                self.post_install = ()
            case _:
                raise ValueError(
                    f"Invalid post_install type: {type(self.post_install)}, expected tuple!"
                )

    @classmethod
    def empty(cls) -> Env:
        """Create an empty environment instance."""
        return cls(name="empty")

    @classmethod
    def merge(cls, one: Env, two: Env) -> Env:
        """Merges two Env instances into one.

        If the python versions of the two environments differ, the lowest version is chosen
        and a warning is issued. The resulting environment will have a combination of the
        requirements and post_install scripts from both environments.

        Args:
            one: The first environment to merge.
            two: The second environment to merge.

        Returns:
            A new environment instance that is the result of merging the two input environments.
        """
        if one.python_version != two.python_version:
            this = tuple(map(int, one.python_version.split(".")))
            that = tuple(map(int, two.python_version.split(".")))
            resolved = min(this, that)
            warnings.warn(
                f"Different python versions for:"
                f" {this} - {one.name}\n"
                f" {that} - {two.name}\n"
                f"Resolving to the lowest version {resolved}.",
                UserWarning,
                stacklevel=2,
            )
            python_version = ".".join(map(str, resolved))
        else:
            python_version = one.python_version

        for package in one.requirements:
            for other_package in two.requirements:
                if not env_pkg_version_compat(package, other_package):
                    raise ValueError(
                        f"Package versions are not compatible: {package}, {other_package}"
                    )


        if one.name == two.name:
            return Env(
                name=one.name,
                python_version=python_version,
                requirements=tuple(set(one.requirements + two.requirements)),
                post_install=one.post_install + two.post_install,
            )

        return Env(
            name=one.name + "_" + two.name,
            python_version=python_version,
            requirements=tuple(set(one.requirements + two.requirements)),
            post_install=one.post_install + two.post_install,
        )

    @property
    def identifier(self) -> str:
        """Returns the name of the current environment."""
        return self.name

    def to_dict(self) -> dict[str, Any]:
        """Converts the environment configuration to a dictionary."""
        return {
            "name": self.name,
            "python_version": self.python_version,
            "requirements": list(self.requirements),
            "post_install": list(self.post_install),
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Env:
        """Creates an environment configuration from a dictionary."""
        return cls(
            name=data["name"],
            python_version=data["python_version"],
            requirements=tuple(data["requirements"]),
            post_install=tuple(data["post_install"]),
        )


def _get_current_installed_python_version() -> str:
    import sys

    return f"{sys.version_info.major}.{sys.version_info.minor}"


def get_current_installed_hpoglue_version() -> str:
    """Retrieve the currently installed version of hpoglue."""
    cmd = ["pip", "show", "hpoglue"]
    logger.debug(cmd)
    output = subprocess.run(  # noqa: S603
        cmd,
        check=True,
        capture_output=True,
        text=True,
    )
    lines = output.stdout.strip().splitlines()
    for line in lines:
        if "Version: " in line:
            return line.split(": ")[1]

    raise RuntimeError(f"Could not find hpoglue version in {lines}.")


@dataclass
class Venv:
    """Represents a Venv configuration with a path, activate, python, and pip commands."""
    def __init__(self, path: Path) -> None:
        """Initialize the environment with the given virtual environment path."""
        self.venv_path = path
        self.activate = f"{self.venv_path / 'bin' / 'activate'}"
        self.python = f"{self.venv_path / 'bin' / 'python'}"
        self.pip = f"{self.venv_path / 'bin' / 'pip'}"

    def create(
        self,
        path: Path,
        *,
        python_version: str,
        requirements_file: Path,
        exists_ok: bool = False,
    ) -> None:
        """Create a virtual environment at the specified path."""
        if self.venv_path.exists() and not exists_ok:
            raise FileExistsError(f"Virtual environment already exists at {self.venv_path}")

        if self.venv_path.exists():
            return

        current_python_version = _get_current_installed_python_version()
        if current_python_version != python_version:
            raise RuntimeError(
                f"Python version between current env and required env:"
                f" {current_python_version} != {python_version}."
                " Please use some other method of creating virtual environments"
                " that support python versions, for example `'conda'`."
            )

        path = self.venv_path.resolve().absolute()
        cmd = ["python", "-m", "venv", str(path)]
        logger.debug(cmd)
        subprocess.run(cmd, check=True)  # noqa: S603

        if requirements_file.exists():
            cmd = [self.pip, "install", "-r", str(requirements_file)]
            logger.debug(cmd)
            subprocess.run(cmd, check=True)  # noqa: S603

    def run(self, cmd: Sequence[str]) -> None:
        """Run the given command in the virtual environment."""
        cmd = [self.python, "-m", *cmd]
        logger.debug(cmd)
        subprocess.run(cmd, check=True)  # noqa: S603



from __future__ import annotations

import logging
import warnings
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

from hpoglue.budget import CostBudget, TrialBudget
from hpoglue.config import PRECISION, Config
from hpoglue.fidelity import Fidelity, ListFidelity, RangeFidelity
from hpoglue.measure import Measure
from hpoglue.optimizer import Optimizer
from hpoglue.query import Query
from hpoglue.result import Result
from hpoglue.utils import configpriors_to_dict, dict_to_configpriors, first, first_n, mix_n

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from hpoglue.benchmark import BenchmarkDescription
    from hpoglue.budget import BudgetType

logger = logging.getLogger(__name__)

OptWithHps: TypeAlias = tuple[type[Optimizer], Mapping[str, Any]]


@dataclass(kw_only=True, unsafe_hash=True)
class Problem:
    """A problem to optimize over."""

    # NOTE: These are mainly for consumers who need to interact beyond forward facing API
    Config: TypeAlias = Config
    Query: TypeAlias = Query
    Result: TypeAlias = Result
    Measure: TypeAlias = Measure
    TrialBudget: TypeAlias = TrialBudget
    CostBudget: TypeAlias = CostBudget
    RangeFidelity: TypeAlias = RangeFidelity
    ListFidelity: TypeAlias = ListFidelity

    objectives: tuple[str, Measure] | Mapping[str, Measure] = field(hash=False)
    """The metrics to optimize for this problem, with a specific order.

    If only one metric is specified, this is considered single objective and
    not multiobjective.
    """

    fidelities: tuple[str, Fidelity] | Mapping[str, Fidelity] | None = field(hash=False)
    """Fidelities to use from the Benchmark.

    When `None`, the problem is considered a black-box problem with no fidelity.

    When a single fidelity is specified, the problem is considered a _multi-fidelity_ problem.

    When many fidelities are specified, the problem is considered a _many-fidelity_ problem.
    """

    costs: tuple[str, Measure] | Mapping[str, Measure] | None = field(hash=False)
    """The cost metrics to use for this problem.

    When `None`, the problem is considered a black-box problem with no cost.

    When a single cost is specified, the problem is considered a _cost-sensitive_ problem.

    When many costs are specified, the problem is considered a _multi-cost_ problem.
    """

    budget: BudgetType
    """The type of budget to use for the optimizer."""

    optimizer: type[Optimizer]
    """The optimizer to use for this problem"""

    optimizer_hyperparameters: Mapping[str, int | float] = field(default_factory=dict)
    """The hyperparameters to use for the optimizer"""

    benchmark: BenchmarkDescription
    """The benchmark to use for this problem"""

    config_space: ConfigurationSpace | list[Config] = field(init=False)
    """The configuration space for the problem"""

    is_tabular: bool = field(init=False)
    """Whether the benchmark is tabular"""

    is_multiobjective: bool = field(init=False)
    """Whether the problem has multiple objectives"""

    is_multifidelity: bool = field(init=False)
    """Whether the problem has a fidelity parameter"""

    is_manyfidelity: bool = field(init=False)
    """Whether the problem has many fidelities"""

    supports_trajectory: bool = field(init=False)
    """Whether the problem setup allows for trajectories to be queried."""

    name: str = field(init=False)
    """The name of the problem.
        This is used to identify the problem.
    """

    precision: int = field(default=12)
    """The precision to use for the problem."""

    mem_req_mb: int = field(init=False)
    """The memory requirement in MB for the problem.
        It is the sum of the optimizer and benchmark memory requirements.
    """

    continuations: bool = field(default=True)
    """Whether the problem supports continuations."""

    priors: tuple[str, Mapping[str, Config]] = field(default_factory=dict)
    """Priors to use for the problem's objectives.
        Format: (unique_prior_id, {objective_name: prior Config})
    """

    def __post_init__(self) -> None:  # noqa: C901, PLR0912, PLR0915
        self.config_space = self.benchmark.config_space
        self.mem_req_mb = self.optimizer.mem_req_mb + self.benchmark.mem_req_mb
        self.is_tabular = self.benchmark.is_tabular
        self.is_manyfidelity: bool
        self.is_multifidelity: bool
        self.supports_trajectory: bool

        name_parts: list[str] = [
            f"optimizer={self.optimizer.name}",
            f"benchmark={self.benchmark.name}",
            "objectives=" + (
                ",".join(self.get_objectives())
                if isinstance(self.objectives, Mapping)
                else self.get_objectives()
            )
        ]

        if len(self.optimizer_hyperparameters) > 0:
            name_parts.insert(
                1, ",".join(f"{k}={v}" for k, v in self.optimizer_hyperparameters.items())
            )

        if self.fidelities is not None and len(self.benchmark.fidelities) > 1:
            name_parts.append(
                "fidelities=" + (
                    ",".join(self.get_fidelities())
                    if isinstance(self.fidelities, Mapping)
                    else self.get_fidelities()
                )
            )

        if self.costs is not None:
            name_parts.append(
                "costs=" + (
                    ",".join(self.get_costs())
                    if isinstance(self.costs, Mapping)
                    else self.get_costs()
                )
            )

        name_parts.append(self.budget.path_str)

        if self.priors:
            name_parts.append(
                f"priors={self.priors[0]}"
            )

        self.name = ".".join(name_parts)

        self.is_multiobjective: bool
        match self.objectives:
            case tuple():
                self.is_multiobjective = False
            case Mapping():
                if len(self.objectives) == 1:
                    raise ValueError("Single objective should be a tuple, not a mapping")

                self.is_multiobjective = True
            case _:
                raise TypeError("Objectives must be a tuple (name, measure) or a mapping")

        match self.fidelities:
            case None:
                self.is_multifidelity = False
                self.is_manyfidelity = False
                self.supports_trajectory = False
            case (_name, _fidelity):
                self.is_multifidelity = True
                self.is_manyfidelity = False
                if _fidelity.supports_continuation:
                    self.supports_trajectory = True
                else:
                    self.supports_trajectory = False
            case Mapping():
                if len(self.fidelities) == 1:
                    raise ValueError("Single fidelity should be a tuple, not a mapping")

                self.is_multifidelity = False
                self.is_manyfidelity = True
                self.supports_trajectory = False
            case _:
                raise TypeError("Fidelity must be a tuple (name, fidelity) or a mapping")

        match self.costs:
            case None:
                pass
            case (_name, _measure):
                pass
            case Mapping():
                if len(self.costs) == 1:
                    raise ValueError("Single cost should be a tuple, not a mapping")

    @classmethod
    def problem(  # noqa: C901, PLR0912, PLR0913, PLR0915
        cls,
        *,
        optimizer: type[Optimizer],
        optimizer_hyperparameters: Mapping[str, int | float] = {},
        benchmark: BenchmarkDescription,
        budget: BudgetType | int | float,
        fidelities: int | str | list[str] | None = None,
        objectives: int | str | list[str] = 1,
        costs: int = 0,
        multi_objective_generation: Literal["mix_metric_cost", "metric_only"] = "mix_metric_cost",
        precision: int | None = None,
        continuations: bool = True,
        priors: tuple[str, Mapping[str, Config] | Mapping[str, dict[str, Any]]] | None = None,
    ) -> Problem:
        """Generate a problem for this optimizer and benchmark.

        Args:
            optimizer: The optimizer to use for the problem.

            optimizer_hyperparameters: The hyperparameters to use for the optimizer.

            benchmark: The benchmark to use for the problem.

            budget: The budget to use for the problems. Budget defaults to a n_trials budget
                where when multifidelty is enabled, fractional budget can be used and 1 is
                equivalent a full fidelity trial.

            fidelities: The actual fidelities or number of fidelities for the problem.

            objectives: The actual objectives or number of objectives for the problem.

            costs: The number of costs for the problem.

            multi_objective_generation: The method to generate multiple objectives.

            precision: The precision to use for the problem.

            continuations: Whether to use continuations for the problem.

            priors: Priors to use for the problem's objectives.
        """
        _fid: tuple[str, Fidelity] | Mapping[str, Fidelity] | None
        match fidelities:
            case int() if fidelities < 0:
                raise ValueError(f"{fidelities=} must be >= 0")
            case 0:
                _fid = None
            case None:
                _fid = None
            case 1:
                if benchmark.fidelities is None:
                    raise ValueError(
                        (
                            f"Benchmark {benchmark.name} has no fidelities but {fidelities=} "
                            "was requested"
                        ),
                    )
                _fid = first(benchmark.fidelities)
            case int():
                if benchmark.fidelities is None:
                    raise ValueError(
                        (
                            f"Benchmark {benchmark.name} has no fidelities but {fidelities=} "
                            "was requested"
                        ),
                    )

                if fidelities > len(benchmark.fidelities):
                    raise ValueError(
                        f"{fidelities=} is greater than the number of fidelities"
                        f" in benchmark {benchmark.name} which has "
                        f"{len(benchmark.fidelities)} fidelities",
                    )

                _fid = first_n(fidelities, benchmark.fidelities)
            case str():
                if benchmark.fidelities is None:
                    raise ValueError(
                        (
                            f"Benchmark {benchmark.name} has no fidelities but {fidelities=} "
                            "was requested"
                        ),
                    )
                if fidelities not in benchmark.fidelities:
                    raise ValueError(
                        f"{fidelities=} not found in benchmark {benchmark.name} fidelities",
                    )
                _fid = (fidelities, benchmark.fidelities[fidelities])
            case list():
                if benchmark.fidelities is None:
                    raise ValueError(
                        (
                            f"Benchmark {benchmark.name} has no fidelities but {fidelities=} "
                            "was requested"
                        ),
                    )
                if len(fidelities) > len(benchmark.fidelities):
                    raise ValueError(
                        f"{fidelities=} is greater than the number of fidelities"
                        f" in benchmark {benchmark.name} which has "
                        f"{len(benchmark.fidelities)} fidelities",
                    )
                _fid = {name: benchmark.fidelities[name] for name in fidelities}
            case _:
                raise TypeError(f"{fidelities=} not supported")

        _obj: tuple[str, Measure] | Mapping[str, Measure]
        match objectives, multi_objective_generation:
            # single objective
            case int(), _ if objectives < 0:  # type: ignore
                raise ValueError(f"{objectives=} must be >= 0")
            case _, str() if multi_objective_generation not in {"mix_metric_cost", "metric_only"}:
                raise ValueError(
                    f"{multi_objective_generation=} not supported, must be one"
                    " of 'mix_metric_cost', 'metric_only'",
                )
            case 1, _:
                _obj = first(benchmark.metrics)
            case int(), "metric_only":
                if objectives > len(benchmark.metrics):  # type: ignore
                    raise ValueError(
                        f"{objectives=} is greater than the number of metrics"
                        f" in benchmark {benchmark.name} which has {len(benchmark.metrics)} metrics"
                    )
                _obj = first_n(objectives, benchmark.metrics)  # type: ignore
            case int(), "mix_metric_cost":
                n_costs = 0 if benchmark.costs is None else len(benchmark.costs)
                n_available = len(benchmark.metrics) + n_costs
                if objectives > n_available:  # type: ignore
                    raise ValueError(
                        f"{objectives=} is greater than the number of metrics and costs"
                        f" in benchmark {benchmark.name} which has {n_available} objectives"
                        " when combining metrics and costs",
                    )
                if benchmark.costs is None:
                    _obj = first_n(objectives, benchmark.metrics)  # type: ignore
                else:
                    _obj = mix_n(objectives, benchmark.metrics, benchmark.costs)  # type: ignore
            case str(), _:
                if objectives not in benchmark.metrics:
                    raise ValueError(
                        f"{objectives=} not found in benchmark {benchmark.name} metrics",
                    )
                _obj = (objectives, benchmark.metrics[objectives])  # type: ignore
            case list(), "metric_only":
                if len(objectives) > len(benchmark.metrics):  # type: ignore
                    raise ValueError(
                        f"{objectives=} is greater than the number of metrics"
                        f" in benchmark {benchmark.name} which has {len(benchmark.metrics)} metrics"
                    )
                _obj = {name: benchmark.metrics[name] for name in objectives}  # type: ignore
            case list(), "mix_metric_cost":
                n_costs = 0 if benchmark.costs is None else len(benchmark.costs)
                n_available = len(benchmark.metrics) + n_costs
                if len(objectives) > n_available:  # type: ignore
                    raise ValueError(
                        f"{objectives=} is greater than the number of metrics and costs"
                        f" in benchmark {benchmark.name} which has {n_available} objectives"
                        " when combining metrics and costs",
                    )
                if benchmark.costs is None:
                    for obj in objectives:  # type: ignore
                        if obj not in benchmark.metrics:
                            raise ValueError(
                                f"{obj=} not found in benchmark {benchmark.name} metrics",
                            )
                    _obj = {name: benchmark.metrics[name] for name in objectives}  # type: ignore
                else:
                    _obj = mix_n(len(objectives), benchmark.metrics, benchmark.costs)  # type: ignore
            case _, _:
                raise RuntimeError(
                    f"Unexpected case with {objectives=}, {multi_objective_generation=}",
                )

        _cost: tuple[str, Measure] | Mapping[str, Measure] | None
        match costs:
            case int() if costs < 0:
                raise ValueError(f"{costs=} must be >= 0")
            case 0:
                _cost = None
            case None:
                _cost = None
            case 1:
                if benchmark.costs is None:
                    raise ValueError(
                        f"Benchmark {benchmark.name} has no costs but {costs=} was requested",
                    )
                _cost = first(benchmark.costs)
            case int():
                if benchmark.costs is None:
                    raise ValueError(
                        f"Benchmark {benchmark.name} has no costs but {costs=} was requested",
                    )
                _cost = first_n(costs, benchmark.costs)
            case _:
                raise TypeError(f"{costs=} not supported")

        _budget: BudgetType
        match budget:
            case int() if budget < 0:
                raise ValueError(f"{budget=} must be >= 0")
            case int():
                _budget = TrialBudget(budget)
            case float():
                _budget = CostBudget(budget)
            case TrialBudget():
                _budget = budget
            case CostBudget():
                raise NotImplementedError("Cost budgets are not yet supported")
            case _:
                raise TypeError(f"Unexpected type for `{budget=}`: {type(budget)}")

        _opt = optimizer[0] if isinstance(optimizer, tuple) else optimizer

        match priors:
            case None:
                pass
            case tuple():
                priors = dict_to_configpriors(priors)
            case _:
                raise TypeError(f"Unexpected type for priors: {type(priors)}")

        if "single" not in _opt.support.fidelities:
            continuations = False

        problem = Problem(
            optimizer=optimizer,
            optimizer_hyperparameters=optimizer_hyperparameters,
            benchmark=benchmark,
            budget=_budget,
            fidelities=_fid,
            objectives=_obj,
            costs=_cost,
            precision=precision if precision is not None else PRECISION,
            continuations=continuations,
            priors=priors,
        )

        support: Problem.Support = _opt.support
        support.check_opt_support(who=_opt.name, problem=problem)

        return problem

    def get_objectives(self) -> str | list[str]:
        """Retrieve the objectives of the problem.

        Returns:
            objectives of the problem
        """
        return (
            list(self.objectives.keys())
            if isinstance(self.objectives, Mapping)
            else self.objectives[0]
        )

    def get_fidelities(self) -> str | list[str] | None:
        """Retrieve the fidelities associated with the object.

        Returns:
            fidelities of the problem.
        """
        return (
            None
            if self.fidelities is None
            else list(self.fidelities.keys())
            if isinstance(self.fidelities, Mapping)
            else self.fidelities[0]
        )

    def get_costs(self) -> str | list[str] | None:
        """Retrieve the costs associated with the object.

        Returns:
            costs of the problem.
        """
        return (
            None
            if self.costs is None
            else list(self.costs.keys())
            if isinstance(self.costs, Mapping)
            else self.costs[0]
        )

    def group_for_optimizer_comparison(
        self,
    ) -> tuple[
        str,
        BudgetType,
        tuple[tuple[str, Measure], ...],
        None | tuple[tuple[str, Fidelity], ...],
        None | tuple[tuple[str, Measure], ...],
    ]:
        """Groups the objectives, fidelities, and costs for optimizer comparison.

        Returns:
            tuple: A tuple containing:
                - The name of the benchmark.
                - The budget type.
                - The objectives.
                - The fidelities.
                - The costs.
        """
        match self.objectives:
            case (name, measure):
                _obj = ((name, measure),)
            case Mapping():
                _obj = tuple(self.objectives.items())

        match self.fidelities:
            case None:
                _fid = None
            case (name, fid):
                _fid = ((name, fid),)
            case Mapping():
                _fid = tuple(self.fidelities.items())

        match self.costs:
            case None:
                _cost = None
            case (name, measure):
                _cost = ((name, measure),)
            case Mapping():
                _cost = tuple(self.costs.items())

        return (self.benchmark.name, self.budget, _obj, _fid, _cost)

    def to_dict(self) -> dict[str, Any]:
        """Convert the problem instance to a dictionary."""
        return {
            "objectives": self.get_objectives(),
            "fidelities": self.get_fidelities(),
            "costs": self.get_costs(),
            "budget_type": self.budget.name,
            "budget": self.budget.to_dict(),
            "benchmark": self.benchmark.name,
            "optimizer": self.optimizer.name,
            "optimizer_hyperparameters": self.optimizer_hyperparameters,
            "continuations": self.continuations,
            "priors": configpriors_to_dict(self.priors) if self.priors else None,
        }

    @classmethod
    def from_dict(  # noqa: C901, PLR0912
        cls,
        data: dict[str, Any],
        benchmarks_dict: Mapping[str, BenchmarkDescription],
        optimizers_dict: Mapping[str, Optimizer],
    ) -> Problem:
        """Create a Problem instance from a dictionary.

        Args:
            data: A dictionary containing the problem data.
            benchmarks_dict: A mapping of benchmark names to BenchmarkDescription instances.
            optimizers_dict: A mapping of optimizer names to Optimizer instances.

        Returns:
            A Problem instance created from the dictionary data.
        """
        if data["benchmark"] not in benchmarks_dict:
            raise ValueError(
                f"Benchmark {data['benchmark']} not found in benchmarks!"
                " Please make sure your benchmark is registed in `BENCHMARKS`"
                " before loading/parsing."
            )

        if data["optimizer"] not in optimizers_dict:
            raise ValueError(
                f"Optimizer {data['optimizer']} not found in optimizers!"
                " Please make sure your optimizer is registed in `OPTIMIZERS`"
                " before loading/parsing."
            )

        benchmark = benchmarks_dict[data["benchmark"]]
        optimizer = optimizers_dict[data["optimizer"]]
        _obj = data["objectives"]
        match _obj:
            case str():
                objectives = (_obj, benchmark.metrics[_obj])
            case list():
                objectives = {name: benchmark.metrics[name] for name in _obj}
            case _:
                raise ValueError("Objectives must be a string or a list of strings")

        _fid = data["fidelities"]
        match _fid:
            case None:
                fidelities = None
            case str():
                assert benchmark.fidelities is not None
                fidelities = (_fid, benchmark.fidelities[_fid])
            case list():
                assert benchmark.fidelities is not None
                fidelities = {name: benchmark.fidelities[name] for name in _fid}
            case _:
                raise ValueError("Fidelity must be a string or a list of strings")

        _cost = data["costs"]
        match _cost:
            case None:
                costs = None
            case str():
                assert benchmark.costs is not None
                costs = (_cost, benchmark.costs[_cost])
            case list():
                assert benchmark.costs is not None
                costs = {name: benchmark.costs[name] for name in _cost}
            case _:
                raise ValueError("Costs must be a string or a list of strings")

        _budget_type = data["budget_type"]
        match _budget_type:
            case "trial_budget":
                budget = TrialBudget.from_dict(data["budget"])
            case "cost_budget":
                budget = CostBudget.from_dict(data["budget"])
            case _:
                raise ValueError("Budget type must be 'trial_budget' or 'cost_budget'")

        return cls(
            objectives=objectives,
            fidelities=fidelities,
            costs=costs,
            budget=budget,
            benchmark=benchmark,
            optimizer=optimizer,
            optimizer_hyperparameters=data["optimizer_hyperparameters"],
            continuations=data["continuations"],
            priors = dict_to_configpriors(data["priors"]) if data.get("priors") else None,
        )

    @dataclass(kw_only=True)
    class Support:
        """The support of an optimizer for a problem."""

        objectives: tuple[Literal["single", "many"], ...] = field(default=("single",))
        fidelities: tuple[Literal[None, "single", "many"], ...] = field(default=(None,))
        cost_awareness: tuple[Literal[None, "single", "many"], ...] = field(default=(None,))
        tabular: bool = False
        priors: bool = False

        def __post_init__(self) -> None:
            match self.objectives:
                case tuple():
                    pass
                case str():
                    self.objectives = (self.objectives,)
                case _:
                    raise ValueError(
                        "Invalid type for optimizer support objectives: "
                        f"{type(self.objectives)}, expected tuple!"
                    )
            match self.fidelities:
                case tuple():
                    pass
                case str():
                    self.fidelities = (self.fidelities,)
                case None:
                    self.fidelities = (None,)
                case _:
                    raise ValueError(
                        "Invalid type for optimizer support fidelities: "
                        f"{type(self.fidelities)}, expected tuple!"
                    )
            match self.cost_awareness:
                case tuple():
                    pass
                case str():
                    self.cost_awareness = (self.cost_awareness,)
                case None:
                    self.cost_awareness = (None,)
                case _:
                    raise ValueError(
                        "Invalid type for optimizer support cost_awareness: "
                        f"{type(self.cost_awareness)}, expected tuple!"
                    )

        def check_opt_support(self, who: str, *, problem: Problem) -> None:  # noqa: C901
            """Check if the problem is supported by the support."""
            match problem.fidelities:
                case None if None not in self.fidelities:
                    raise ValueError(
                        f"Optimizer {who} does not support having no fidelties for {problem.name}!"
                    )
                case tuple() if "single" not in self.fidelities:
                    raise ValueError(
                        f"Optimizer {who} does not support multi-fidelty for {problem.name}!"
                    )
                case Mapping() if "many" not in self.fidelities:
                    raise ValueError(
                        f"Optimizer {who} does not support many-fidelty for {problem.name}!"
                    )

            match problem.objectives:
                case tuple() if "single" not in self.objectives:
                    raise ValueError(
                        f"Optimizer {who} does not support single-objective for {problem.name}!"
                    )
                case Mapping() if "many" not in self.objectives:
                    raise ValueError(
                        f"Optimizer {who} does not support multi-objective for {problem.name}!"
                    )

            match problem.costs:
                case None if None not in self.cost_awareness:
                    raise ValueError(
                        f"Optimizer {who} does not support having no cost for {problem.name}!"
                    )
                case tuple() if "single" not in self.cost_awareness:
                    raise ValueError(
                        f"Optimizer {who} does not support single-cost for {problem.name}!"
                    )
                case Mapping() if "many" not in self.cost_awareness:
                    raise ValueError(
                        f"Optimizer {who} does not support multi-cost for {problem.name}!"
                    )

            match problem.is_tabular:
                case True if not self.tabular:
                    raise ValueError(
                        f"Optimizer {who} does not support tabular benchmarks for {problem.name}!"
                    )

            if problem.priors and not self.priors:
                warnings.warn(
                    f"Optimizer {who} does not support priors",
                    stacklevel=2,
                )
                problem.priors = None
                problem.name = problem.name.split("priors=")[0]



from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any, TypeAlias, TypeVar

import numpy as np
import pandas as pd
from more_itertools import roundrobin, take
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from hpoglue.config import Config

logger = logging.getLogger(__name__)

DataContainer: TypeAlias = np.ndarray | (pd.DataFrame | pd.Series)
D = TypeVar("D", bound=DataContainer)


def scale(
    unit_xs: int | float | np.number | np.ndarray | pd.Series,
    to: tuple[int | float | np.number, int | float | np.number],
) -> float | np.number | np.ndarray | pd.Series:
    """Scale values from unit range to a new range.

    >>> scale(np.array([0.0, 0.5, 1.0]), to=(0, 10))
    array([ 0.,  5., 10.])

    Parameters
    ----------
    unit_xs:
        The values to scale

    to:
        The new range

    Returns:
    -------
        The scaled values
    """
    return unit_xs * (to[1] - to[0]) + to[0]  # type: ignore


def normalize(
    x: int | float | np.number | np.ndarray | pd.Series,
    *,
    bounds: tuple[int | float | np.number, int | float | np.number],
) -> float | np.number | np.ndarray | pd.Series:
    """Normalize values to the unit range.

    >>> normalize(np.array([0.0, 5.0, 10.0]), bounds=(0, 10))
    array([0. , 0.5, 1. ])

    Parameters
    ----------
    x:
        The values to normalize

    bounds:
        The bounds of the range

    Returns:
    -------
        The normalized values
    """
    if bounds == (0, 1):
        return x

    return (x - bounds[0]) / (bounds[1] - bounds[0])  # type: ignore


def rescale(
    x: int | float | np.number | np.ndarray | pd.Series,
    frm: tuple[int | float | np.number, int | float | np.number],
    to: tuple[int | float | np.number, int | float | np.number],
) -> float | np.ndarray | pd.Series:
    """Rescale values from one range to another.

    >>> rescale(np.array([0, 10, 20]), frm=(0, 100), to=(0, 10))
    array([0, 1, 2])

    Parameters
    ----------
    x:
        The values to rescale

    frm:
        The original range

    to:
        The new range

    Returns:
    -------
        The rescaled values
    """
    if frm != to:
        normed = normalize(x, bounds=frm)
        scaled = scale(unit_xs=normed, to=to)
    else:
        scaled = x

    match scaled:
        case int() | float() | np.number():
            return float(scaled)
        case np.ndarray() | pd.Series():
            return scaled.astype(np.float64)
        case _:
            raise ValueError(f"Unsupported type {type(x)}")


T = TypeVar("T")

def first(_d: Mapping[str, T]) -> tuple[str, T]:
    """Return the first key-value pair from a dictionary.
        Used for retrieving the first objectives/fidelities/costs of a benchmark.

    Args:
        _d: The dictionary from which to retrieve the first item.

    Returns:
        A tuple containing the first key and its corresponding value from the dictionary.

    Raises:
        StopIteration: If the dictionary is empty.
    """
    return next(iter(_d.items()))


def first_n(n: int, _d: Mapping[str, T]) -> dict[str, T]:
    """Return the first `n` items from the given dictionary.
        Used for retrieving the first `n` objectives/fidelities/costs of a benchmark.

    Args:
        n (int): The number of items to return.
        _d (Mapping[str, T]): The dictionary from which to take the items.

    Returns:
        dict[str, T]: A dictionary containing the first `n` items from `_d`.
    """
    return dict(take(n, _d.items()))


def mix_n(n: int, _d1: Mapping[str, T], _d2: Mapping[str, T]) -> dict[str, T]:
    """Mixes items from two dictionaries in a round-robin fashion up to a specified number of items.
        Used for mixing objectives/costs of benchmark in case of multi-objective optimization.

    Args:
        n (int): The number of items to take from the combined dictionaries.
        _d1 (Mapping[str, T]): The first dictionary to mix.
        _d2 (Mapping[str, T]): The second dictionary to mix.

    Returns:
        dict[str, T]: A dictionary containing up to `n` items from the combined dictionaries.
    """
    return dict(take(n, roundrobin(_d1.items(), _d2.items())))


def configpriors_to_dict(
        priors: tuple[str, Mapping[str, Config]]
    ) -> tuple[str, Mapping[str, Mapping[str, Any]]]:
    """Converts a tuple of priors to a dictionary.

    Args:
        priors: A tuple with the priors as Config objects..

    Returns:
        A tuple with the priors converted to dictionaries.
    """
    assert isinstance(priors, tuple | list) and len(priors) == 2, (  # noqa: PLR2004, PT018
        "Priors should be a tuple or list of length 2 with the format: "
        "(str, Mapping[str, Config])."
    )
    name, _priors = priors
    prior_dict = {}
    for obj, prior in _priors.items():
        match prior:
            case Config():
                prior_dict[obj] = prior.values
            case dict():
                prior_dict[obj] = prior
            case _:
                raise TypeError(f"Unsupported type for priors: {type(prior)}")

    return name, prior_dict


def dict_to_configpriors(
    priors: tuple[str, Mapping[str, Mapping[str, Any]]]
) -> tuple[str, Mapping[str, Config]]:
    """Converts priors given as dictionaries into Config objects.

    Args:
        priors: The tuple of str and dictionary of priors to convert,
        with the priors themselves as dictionaries.

    Returns:
        A tuple with the priors converted into Config objects.
    """
    assert isinstance(priors, tuple | list) and len(priors) == 2, (  # noqa: PLR2004, PT018
        "Priors should be a tuple or list of length 2 with the format: "
        "(str, Mapping[str, dict[str, Any]])."
    )
    name, _priors = priors
    prior_dict = {}
    for obj, prior in _priors.items():
        match prior:
            case dict():
                prior_dict[obj] = Config(
                    config_id=obj,
                    values=prior,
                )
            case Config():
                prior_dict[obj] = prior
            case _:
                raise TypeError(f"Unsupported type for priors: {type(prior)}")

    return name, prior_dict


def env_pkg_version_compat(
    package1: str,
    package2: str,
):
    """Check if two package versions are compatible.

    Args:
        package1: The first package version.
        package2: The second package version.

    Returns:
        bool: True if the package versions are compatible, False otherwise.
    """
    name1, key1, spec1 = _split_pkg_ver(package1)
    name2, key2, spec2 = _split_pkg_ver(package2)

    if name1 != name2:
        return True

    if not spec1 or not spec2:
        return True

    key_spec1, key_spec2 = SpecifierSet(key1 + spec1), SpecifierSet(key2 + spec2)

    return bool(key_spec1.contains(Version(spec2)) or key_spec2.contains(Version(spec1)))


def _split_pkg_ver(package_name: str) -> tuple[str, str, str]:
    """Split package name into name, operator, and version.

    Args:
        package_name: The package name to split.

    Returns:
        tuple: A tuple containing the package name and specifier set.
    """
    version_constraints = ["==", ">=", "<=", ">", "<"]
    for constraint in version_constraints:
        if constraint in package_name:
            package_name, version_spec = package_name.split(constraint)
            return package_name, constraint, version_spec

    return package_name, "", ""


from __future__ import annotations

import logging
import warnings
from collections.abc import Mapping
from contextlib import nullcontext
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
from tqdm import TqdmWarning, tqdm

from hpoglue.budget import CostBudget, TrialBudget
from hpoglue.fidelity import Fidelity

if TYPE_CHECKING:
    from hpoglue.problem import Problem
from hpoglue.utils import rescale

if TYPE_CHECKING:
    from hpoglue.benchmark import Benchmark
    from hpoglue.optimizer import Optimizer
    from hpoglue.result import Result

logger = logging.getLogger(__name__)


# TODO(eddiebergman):
# 1. Do we include the results of everything in between?
#    Would be good to include it in the actual results, if wanting plot the learning
#    curve of individual configurations, not just at the gven fidelity point
# 2. Some optimizers prefer to have some kind of interuption mechanism, i.e.
# > config = opt.ask()
# > for step in steps:
# >     result = benchmark.query(config)
# >   decision = opt.tell(result)
# >   if decision == "stop":
# >         break
@dataclass
class Conf:
    t: tuple
    fid: int | float


@dataclass
class RuntimeHist:
    configs: dict[tuple, dict[str, list[int | float]]] = field(default_factory=dict)

    def add_conf(self, config: Conf, fid_name: str) -> bool:
        if config.t not in self.configs:
            self.configs[config.t] = {fid_name: [config.fid]}
            return False

        if fid_name not in self.configs[config.t]:
            self.configs[config.t][fid_name] = [config.fid]
            return False

        if config.fid in self.configs[config.t][fid_name]:
            warnings.warn(
                f"Fidelity {config.fid} sampled twice by Optimizer for config {config.t}!",
                stacklevel=2,
            )
            return True

        if config.fid < self.configs[config.t][fid_name][-1]:
            raise NotImplementedError("Decreasing fidelity not yet implemented!")

        if config.fid not in self.configs[config.t][fid_name]:
            self.configs[config.t][fid_name].append(config.fid)
            return False

        raise NotImplementedError("Unknown error in adding configuration to history!")

    def get_continuations_cost(self, config: Conf, fid_name: str) -> float:
        fid_list = self.configs[config.t][fid_name]
        if len(fid_list) == 1:
            return fid_list[0]
        return fid_list[-1] - fid_list[-2]

    def get_conf_dict(self) -> dict:
        return self.configs


def _run(
    problem: Problem,
    seed: int,
    *,
    run_name: str | None = None,
    on_error: Literal["raise", "continue"] = "raise",
    progress_bar: bool = False,
) -> list[Result]:
    run_name = run_name if run_name is not None else problem.name
    benchmark = problem.benchmark.load(problem.benchmark)
    opt = problem.optimizer(
        problem=problem,
        working_directory=Path("./Optimizers_cache"),
        seed=seed,
        **problem.optimizer_hyperparameters,
    )

    match problem.budget:
        case TrialBudget(
            total=budget_total,
            minimum_fidelity_normalized_value=minimum_normalized_fidelity,
        ):
            history = _run_problem_with_trial_budget(
                run_name=run_name,
                optimizer=opt,
                benchmark=benchmark,
                problem=problem,
                budget_total=budget_total,
                on_error=on_error,
                minimum_normalized_fidelity=minimum_normalized_fidelity,
                progress_bar=progress_bar,
            )
        case CostBudget():
            raise NotImplementedError("CostBudget not yet implemented")
        case _:
            raise RuntimeError(f"Invalid budget type: {problem.budget}")

    logger.info(f"COMPLETED running {run_name}")
    return history


def _run_problem_with_trial_budget(  # noqa: C901, PLR0912, PLR0915
    *,
    run_name: str,
    optimizer: Optimizer,
    benchmark: Benchmark,
    problem: Problem,
    budget_total: int,
    on_error: Literal["raise", "continue"],
    minimum_normalized_fidelity: float,
    progress_bar: bool,
) -> list[Result]:
    used_budget: float = 0.0

    history: list[Result] = []

    if progress_bar:
        ctx = partial(tqdm, desc=f"{run_name}", total=budget_total)
    else:
        ctx = partial(nullcontext, None)

    # NOTE(eddiebergman): Ignore the tqdm warning about the progress bar going past max
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=TqdmWarning)
        runhist = RuntimeHist()

        with ctx() as pbar:
            while used_budget < budget_total:
                try:
                    query = optimizer.ask()

                    match query.fidelity:
                        case None:
                            result = benchmark.query(query)
                        case Mapping():
                            raise NotImplementedError("Manyfidelity not yet implemented")
                        case (fid_name, fid_value):
                            config = Conf(query.config.to_tuple(problem.precision), fid_value)
                            exist_already = runhist.add_conf(config=config, fid_name=fid_name)

                            if exist_already:
                                # NOTE: Not a cheap operation since we don't store the costs
                                # in the continuations dict
                                result = None
                                for existing_result in history:
                                    tup = existing_result.config.to_tuple(problem.precision)
                                    if Conf(tup, fid_value) == config:
                                        if query.config_id == existing_result.query.config_id:
                                            raise ValueError(
                                                "Resampled configuration has same config_id"
                                                " in history!"
                                            )
                                        existing_result.query = query
                                        result = existing_result
                                        break

                                if result is None:
                                    raise ValueError(
                                        "Resampled configuration not found in history!"
                                    )

                            else:
                                result = benchmark.query(query)
                                if problem.continuations:
                                    result.continuations_cost = runhist.get_continuations_cost(
                                        config=config,
                                        fid_name=fid_name,
                                    )
                        case _:
                            raise TypeError(
                                "Fidelity must be None, tuple(str, value), or Mapping[str, fid]"
                                f" but got: {query.fidelity}"
                            )

                    budget_cost = _trial_budget_cost(
                        value=result.fidelity,
                        problem=problem,
                        minimum_normalized_fidelity=minimum_normalized_fidelity,
                    )

                    used_budget += budget_cost
                    result.budget_cost = budget_cost
                    result.budget_used_total = used_budget

                    optimizer.tell(result)
                    history.append(result)
                    if pbar is not None:
                        pbar.update(budget_cost)

                except Exception as e:
                    logger.exception(e)
                    logger.error(f"Error running {run_name}: {e}")
                    match on_error:
                        case "raise":
                            raise e
                        case "continue":
                            raise NotImplementedError("Continue not yet implemented!") from e
                        case _:
                            raise RuntimeError(f"Invalid value for `on_error`: {on_error}") from e
    return history


def _trial_budget_cost(
    *,
    value: None | tuple[str, int | float] | Mapping[str, int | float],
    problem: Problem,
    minimum_normalized_fidelity: float,
) -> float:
    problem_fids = problem.fidelities
    match value:
        case None:
            assert problem_fids is None
            return 1

        case (name, v):
            assert isinstance(v, int | float | np.integer | np.floating)
            assert isinstance(problem_fids, tuple)
            assert problem_fids[0] == name
            normed_value = rescale(
                v,
                frm=(problem_fids[1].min, problem_fids[1].max),
                to=(minimum_normalized_fidelity, 1),
            )
            assert isinstance(normed_value, float)
            return normed_value

        case Mapping():
            assert isinstance(problem_fids, Mapping)
            assert len(value) == len(problem_fids)
            normed_fidelities: list[float] = []
            for k, v in value.items():
                assert isinstance(v, int | float | np.integer | np.floating)
                assert isinstance(problem_fids[k], Fidelity)
                normed_fid = rescale(
                    v,
                    frm=(problem_fids[k].min, problem_fids[k].max),
                    to=(minimum_normalized_fidelity, 1),
                )
                assert isinstance(normed_fid, float)
                normed_fidelities.append(normed_fid)
            return sum(normed_fidelities) / len(value)
        case _:
            raise TypeError("Fidelity must be None, tuple(str, value), or Mapping[str, fid]")



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    import pandas as pd

    from hpoglue.config import Config
    from hpoglue.query import Query


@dataclass(kw_only=True)
class Result:
    """The result of a query from a benchmark."""

    query: Query
    """The query that generated this result"""

    fidelity: tuple[str, int | float] | Mapping[str, int | float] | None
    """What fidelity the result is at, usually this will be the same as the query fidelity,
    unless the benchmark has multiple fidelities.
    """

    values: dict[str, Any]
    """Everything returned by the benchmark for a given query at the fideltiy."""

    continuations_cost: float = np.nan
    """The coninuations cost if run.continuations set to True."""

    budget_cost: float = np.nan
    """The amount of budget used to generate this result."""

    budget_used_total: float = np.nan
    """The amount of budget used in total."""

    trajectory: pd.DataFrame | None = None
    """If given, the trajectory of the query up to the given fidelity.

    This will only provided if:
    * The problem says it should be provided.
    * There is only a single fidelity parameter.

    It will be multi-indexed, with the multi indexing consiting of the
    config id and the fidelity.
    """

    @property
    def config(self) -> Config:
        """The config."""
        return self.query.config

    def _to_dict(self) -> dict[str, Any]:
        """Convert the result to a dictionary."""
        members = [
            attr for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        members.remove("query")
        members.remove("config")
        members.remove("trajectory")

        _self_dict = {attr: getattr(self, attr) for attr in members}
        _self_dict.update(self.query._to_dict())
        _self_dict["results"] = _self_dict.pop("values")

        return _self_dict


from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Any, Protocol, TypeAlias

import pandas as pd

from hpoglue.config import Config
from hpoglue.env import Env
from hpoglue.optimizer import Optimizer
from hpoglue.result import Result

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from hpoglue.fidelity import Fidelity
    from hpoglue.measure import Measure
    from hpoglue.query import Query

    class TrajectoryF(Protocol):  # noqa: D101
        def __call__(  # noqa: D102
            self,
            *,
            query: Query,
            frm: int | float | None = None,
            to: int | float | None = None,
        ) -> pd.DataFrame: ...


logger = logging.getLogger(__name__)

OptWithHps: TypeAlias = tuple[type[Optimizer], Mapping[str, Any]]


@dataclass(kw_only=True, frozen=True)
class BenchmarkDescription:
    """Describes a benchmark without loading it all in."""

    name: str
    """Unique name of the benchmark."""

    config_space: ConfigurationSpace | list[Config]
    """The configuration space for the benchmark."""

    load: Callable[[BenchmarkDescription], Benchmark] = field(compare=False)
    """Function to load the benchmark."""

    metrics: Mapping[str, Measure]
    """All the metrics that the benchmark supports."""

    test_metrics: Mapping[str, Measure] | None = None
    """All the test metrics that the benchmark supports."""

    costs: Mapping[str, Measure] | None = None
    """All the costs that the benchmark supports."""

    fidelities: Mapping[str, Fidelity] | None = None
    """All the fidelities that the benchmark supports."""

    has_conditionals: bool = False
    """Whether the benchmark has conditionals."""

    is_tabular: bool = False
    """Whether the benchmark is tabular."""

    env: Env = field(default_factory=Env.empty())
    """The environment needed to run this benchmark."""

    mem_req_mb: int = 1024
    """The memory requirement of the benchmark in mb."""

    predefined_points: Mapping[str, Config] | None = None
    """
    Predefined points for the benchmark with their names and descriptions.

    Example:
    ``` python
    {
        "optimum": Config(
                        config_id="optimum",
                        description="This point yields the global optimum",
                        values={"x": 0.0, "y": 0.0}
                    ),
        "worst": Config(
                        config_id="worst",
                        description="This point yields the analytical worst value",
                        values={"x": 1.0, "y": 1.0}
                    ),
    }
    ```
    """

    extra: Mapping[str, Any] = field(default_factory=dict)
    """Extra information about the benchmark."""


@dataclass(kw_only=True)
class SurrogateBenchmark:
    """Defines the interface for a surrogate benchmark."""

    name: str = field(init=False)
    """The name of the Surrogate benchmark."""

    desc: BenchmarkDescription
    """The description of the benchmark."""

    config_space: ConfigurationSpace
    """ The configuration space for the benchmark """

    benchmark: Any
    """The wrapped benchmark object."""

    query: Callable[[Query], Result]
    """The query function for the benchmark."""

    trajectory_f: TrajectoryF | None = None
    """The trajectory function for the benchmark, if one exists.

    This function should return a DataFrame with the trajectory of the query up
    to the given fidelity. The index should be the fidelity parameter with the
    columns as the values.

    ```
    def __call__(
        self,
        *,
        query: Query,
        frm: int | float | None = None,
        to: int | float | None = None,
    ) -> pd.DataFrame:
        ...
    ```

    If not provided, the query will be called repeatedly to generate this.
    """

    def __post_init__(self) -> None:
        self.name = self.desc.name

    def trajectory(  # noqa: D102
        self,
        *,
        query: Query,
        frm: int | float | None = None,
        to: int | float | None = None,
    ) -> pd.DataFrame:
        if self.trajectory_f is not None:
            return self.trajectory_f(query=query, frm=frm, to=to)

        assert isinstance(query.fidelity, tuple)
        assert self.desc.fidelities is not None

        fid_name, fid_value = query.fidelity
        fid = self.desc.fidelities[fid_name]
        frm = frm if frm is not None else fid.min
        to = to if to is not None else fid_value

        index: list[int] | list[float] = []
        results: list[Result] = []
        for val in iter(fid):
            if val < frm:
                continue

            if val > to:
                break

            index.append(val)
            result = self.query(query.with_fidelity((fid_name, val)))
            results.append(result)

        # Return in trajectory format
        # fid_name    **results
        # 0         | . | . | ...
        # 1         | . | . | ...
        # ...
        return pd.DataFrame.from_records(
            [result.values for result in results],
            index=pd.Index(index, name=fid_name),
        )


class TabularBenchmark:
    """Defines the interface for a tabular benchmark."""

    name: str
    """The name of the Tabular benchmark."""

    desc: BenchmarkDescription
    """The description of the benchmark."""

    table: pd.DataFrame
    """ The table holding all information """

    id_key: str
    """The key in the table that we want to use as the id."""

    config_space: list[Config]
    """ All possible configs for the benchmark """

    config_keys: list[str]
    """The keys in the table to use as the config keys."""

    result_keys: list[str]
    """The keys in the table to use as the result keys.

    This is inferred from the `desc=`.
    """

    def __init__(
        self,
        *,
        desc: BenchmarkDescription,
        table: pd.DataFrame,
        id_key: str,
        config_keys: list[str],
    ) -> None:
        """Create a tabular benchmark.

        The result and fidelity keys will be inferred from the `desc=`.

        Args:
            desc: The description of the benchmark.
            table: The table holding all information.
            id_key: The key in the table that we want to use as the id.
            config_keys: The keys in the table that we want to use as the config.
        """
        self.name = desc.name
        # Make sure we work with a clean slate, no issue with index.
        table = table.reset_index()

        for key in config_keys:
            if key not in table.columns:
                raise KeyError(
                    f"Config key '{key}' not in columns {table.columns}."
                    "This is most likely from a misspecified BecnhmarkDescription for "
                    f"{desc.name}.",
                )

        result_keys = [
            *desc.metrics.keys(),
            *(desc.test_metrics.keys() if desc.test_metrics else []),
            *(desc.costs.keys() if desc.costs else []),
        ]
        for key in result_keys:
            if key not in table.columns:
                raise KeyError(
                    f"Result key '{key}' not in columns {table.columns}."
                    "This is most likely from a misspecified BecnhmarkDescription for "
                    f"{desc.name}.",
                )

        match desc.fidelities:
            case None:
                fidelity_keys = None
            case Mapping():
                fidelity_keys = list(desc.fidelities.keys())
                for key in fidelity_keys:
                    if key not in table.columns:
                        raise KeyError(
                            f"Fidelity key '{key}' not in columns {table.columns}."
                            "This is most likely from a misspecified BecnhmarkDescription for "
                            f"{desc.name}.",
                        )
            case _:
                raise TypeError(f"{desc.fidelities=} not supported")

        # Make sure that the column `id` only exist if it's the `id_key`
        if "id" in table.columns and id_key != "id":
            raise ValueError(
                f"Can't have `id` in the columns if it's not the {id_key=}."
                " Please drop it or rename it.",
            )

        # Remap their id column to `id`
        table = table.rename(columns={id_key: "id"})

        # We will create a multi-index for the table, done by the if and
        # the remaining fidelity keys
        _fid_cols = [] if fidelity_keys is None else fidelity_keys

        # Drop all the columns that are not relevant
        relevant_cols: list[str] = ["id", *_fid_cols, *result_keys, *config_keys]
        table = table[relevant_cols]  # type: ignore
        table = table.set_index(["id", *_fid_cols]).sort_index()

        # We now have the following table
        #
        #     id    fidelity | **metrics, **config_values
        #     0         0    |
        #               1    |
        #               2    |
        #     1         0    |
        #               1    |
        #               2    |
        #   ...
        self.table = table
        self.id_key = id_key
        self.desc = desc
        self.config_keys = config_keys
        self.result_keys = result_keys
        self.fidelity_keys = fidelity_keys
        self.config_space = [
            Config(config_id=str(i), values=config)  # enforcing str for id
            for i, config in enumerate(
                self.table[config_keys]
                .drop_duplicates()
                .sort_values(by=config_keys)  # Sorting to ensure table config order is consistent
                .to_dict(orient="records"),
            )
        ]

    def query(self, query: Query) -> Result:
        """Query the benchmark for a result."""
        # NOTE(eddiebergman):
        # Some annoying logic here to basically be able to handle partially specified fidelties,
        # even if it does not match the order of what the table has. In the case where a fidelity
        # is not specified, we select ALL (slice(None)) for that fidelity. Later, we will just take
        # the last row then. Important that we go in the order of `self.fidelity_keys`
        ALL = slice(None)
        fidelity_order = self.table.index.names[1:]

        match query.fidelity:
            case None:
                slices = {col: ALL for col in fidelity_order}
            case (key, value):
                assert self.fidelity_keys is not None
                slices = {col: (value if key == col else ALL) for col in fidelity_order}
            case Mapping():
                assert self.fidelity_keys is not None
                slices = {col: query.fidelity.get(col, ALL) for col in fidelity_order}
            case _:
                raise TypeError(f"type of {query.fidelity=} ({type(query.fidelity)}) supported")

        result = self.table.loc[(query.config_id, *slices.values())]
        row: pd.Series
        match result:
            case pd.Series():
                # If it's a series, a row was uniquely specified, meaning that all
                # of the fidelity values were specified.
                retrieved_results = result[self.result_keys]
                assert isinstance(retrieved_results, pd.Series)
                unspecified_fids = {}
            case pd.DataFrame():
                # If it's a DataFrame, we have multiple rows, we take the last one,
                # under the assumption that:
                # 1. Larger fidelity values are what is requested.
                # 2. The table is sorted by fidelity values.
                retrieved_results = result[self.result_keys]

                # Get the non-specified fidelity values
                # We have to keep it as a dataframe using `[-1:]`
                # for the moment so we can get the correct fidelity names and values.
                row = result.iloc[-1:]
                assert isinstance(row, pd.DataFrame)
                assert len(row.index) == 1
                unspecified_fids = dict(zip(row.index.names, row.index, strict=True))

                retrieved_results = retrieved_results.iloc[-1]
            case _:
                raise TypeError(f"type of {result=} ({type(result)}) not supported")

        match query.fidelity:
            case None:
                fidelities_retrieved = unspecified_fids
            case (key, value):
                fidelities_retrieved = {**unspecified_fids, key: value}
            case Mapping():
                fidelities_retrieved = {**unspecified_fids, **query.fidelity}

        return Result(
            query=query,
            values=retrieved_results.to_dict(),
            fidelity=fidelities_retrieved,
        )

    def trajectory(
        self,
        *,
        query: Query,
        frm: int | float | None = None,
        to: int | float | None = None,
    ) -> pd.DataFrame:
        """Query the benchmark for a result."""
        assert isinstance(query.fidelity, tuple)
        fid_name, fid_value = query.fidelity
        if self.fidelity_keys is None:
            raise ValueError("No fidelities to query for this benchmark!")

        if self.fidelity_keys != [fid_name]:
            raise NotImplementedError(
                f"Can't get trajectory for {fid_name=} when more than one"
                f" fidelity {self.fidelity_keys=}!",
            )

        assert self.desc.fidelities is not None
        frm = frm if frm is not None else self.desc.fidelities[fid_name].min
        to = to if to is not None else fid_value

        # Return in trajectory format
        # fid_name    **results
        # 0         | . | . | ...
        # 1         | . | . | ...
        # ...
        return self.table[self.result_keys].loc[query.config_id, frm:to].droplevel(0).sort_index()


class FunctionalBenchmark:
    """Defines the interface for a functional benchmark."""

    name: str = field(init=False)
    """The name of the Functional benchmark."""

    desc: BenchmarkDescription
    """The description of the functional benchmark."""

    query: Callable[[Query], Result]
    """The query function for the benchmark."""

    config_space: ConfigurationSpace | list[Config] | None = None
    """The configuration space for the benchmark."""

    trajectory_f: TrajectoryF | None = None
    """The trajectory function for the benchmark, if one exists.

    This function should return a DataFrame with the trajectory of the query up
    to the given fidelity. The index should be the fidelity parameter with the
    columns as the values.

    ```
    def __call__(
        self,
        *,
        query: Query,
        frm: int | float | None = None,
        to: int | float | None = None,
    ) -> pd.DataFrame:
        ...
    ```

    If not provided, the query will be called repeatedly to generate this.
    """

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        metrics: Mapping[str, Measure],
        query: Callable[[Query], Result],
        fidelities: Mapping[str, Fidelity] | None = None,
        costs: Mapping[str, Measure] | None = None,
        test_metrics: Mapping[str, Measure] | None = None,
        config_space: ConfigurationSpace | list[Config] | None = None,
        env: Env | None = None,
        mem_req_mb: int = 1024,
        predefined_points: Mapping[str, tuple[Config, str]] | None = None,
        extra: Mapping[str, Any] = {},
    ):
        """Create a functional benchmark.

        Args:
            name: The name of the benchmark.

            metrics: The metrics that the benchmark supports.

            query: The query function for the benchmark.

            fidelities: The fidelities that the benchmark supports.

            costs: The costs that the benchmark supports.

            test_metrics: The test metrics that the benchmark supports.

            config_space: The configuration space for the benchmark.

            env: The environment needed to run this benchmark.

            mem_req_mb: The memory requirement of the benchmark in mb.

            predefined_points: Predefined points for the benchmark with their names and
                                descriptions.

            extra: Extra information about the benchmark.

        """
        self.name = name
        self.query = query
        self.config_space = config_space
        self.desc = BenchmarkDescription(
            name=name,
            config_space=config_space,
            load=partial(self.load),
            metrics=metrics,
            test_metrics=test_metrics,
            costs=costs,
            fidelities=fidelities,
            has_conditionals=False,
            is_tabular=False,
            env=env or Env.empty(),
            mem_req_mb=mem_req_mb,
            predefined_points=predefined_points,
            extra=extra,
        )


    def load(
        self,
        desc: BenchmarkDescription,  # noqa: ARG002
    ) -> FunctionalBenchmark:
        """Load the FunctionalBenchmark."""
        return self


    def trajectory(
        self,
        *,
        query: Query,
        frm: int | float | None = None,
        to: int | float | None = None,
    ) -> pd.DataFrame:
        """The trajectory function for the benchmark, if one exists."""
        if self.trajectory_f is not None:
            return self.trajectory_f(query=query, frm=frm, to=to)

        raise NotImplementedError("Trajectory not implemented for this benchmark.")


# NOTE(eddiebergman): Not using a base class as we really don't expect to need
# more than just these two types of benchmarks.
Benchmark: TypeAlias = TabularBenchmark | SurrogateBenchmark | FunctionalBenchmark



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from hpoglue.env import Env

if TYPE_CHECKING:

    from hpoglue.problem import Problem
    from hpoglue.query import Query
    from hpoglue.result import Result


class Optimizer(ABC):
    """Defines the common interface for Optimizers."""

    name: ClassVar[str]
    """The name of the optimizer"""

    support: ClassVar[Problem.Support]
    """What kind of problems the optimizer supports"""

    multi_fideltiy_requires_learning_curve: ClassVar[bool] = False
    """Whether the optimizer requires a learning curve for multi-fidelity optimization.

    If `True` and the problem is multi-fidelity (1 fidelity) and the fidelity
    supports continuations (e.g. epochs), then a trajectory will be provided in
    the `Result` object provided during the `tell` to the optimizer.
    """

    env: ClassVar[Env] = Env.empty()
    """The environment to setup the optimizer in for `isolated` mode.

    If left as `None`, the currently activated environemnt will be used.
    """

    mem_req_mb: ClassVar[int] = 1024
    """The memory requirement of the optimizer in mb."""

    @abstractmethod
    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        **optimizer_kwargs: Any,
    ) -> None:
        """Initialize the optimizer.

        Args:
            problem: The problem to optimize over
            seed: The random seed for the optimizer
            working_directory: The directory to save the optimizer's state
            optimizer_kwargs: Any additional hyperparameters for the optimizer
        """

    @abstractmethod
    def ask(self) -> Query:
        """Ask the optimizer for a new config to evaluate."""

    @abstractmethod
    def tell(self, result: Result) -> None:
        """Tell the optimizer the result of the query."""




from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import pandas as pd

from hpoglue import Config, FunctionalBenchmark, Problem
from hpoglue._run import _run
from hpoglue.utils import dict_to_configpriors

if TYPE_CHECKING:
    from hpoglue.benchmark import BenchmarkDescription
    from hpoglue.optimizer import Optimizer


def run_glue(  # noqa: C901, PLR0912
    optimizer: type[Optimizer],
    benchmark: BenchmarkDescription | FunctionalBenchmark,
    objectives: int | str | list[str] = 1,
    fidelities: int | str | list[str] | None = None,
    optimizer_hyperparameters: Mapping[str, int | float] | None = None,
    run_name: str | None = None,
    budget: int | float = 50,
    seed: int = 0,
    *,
    continuations: bool = True,
    priors: tuple[str, Mapping[str, Config | Mapping[str, Any]]] | None = None,
) -> pd.DataFrame:
    """Run the glue function using the specified optimizer, benchmark, and hyperparameters.

        optimizer: The optimizer instance to be used.

        benchmark: The benchmark to be evaluated.

        objectives: The objectives for the benchmark.
            Defaults to 1, the first objective in the benchmark.

        fidelities: The fidelities for the benchmark.

        optimizer_hyperparameters: Hyperparameters for the optimizer.

        run_name: An optional name for the run.

        budget: The budget allocated for the run.

        seed: The seed for random number generation to ensure reproducibility.

        continuations: Whether to use continuations for the run.

        priors: Priors to use for the run.

                Advised usage for priors:
                --------------------------

                ```python
                            priors = (
                                        prior_identifier,
                                        {
                                            "objective_name": Config(
                                                config_id="some identifier for this config ID",
                                                values={"hyperparameter name": "prior value"},
                                            )
                                        }
                                    )
                ```
                NOTE:
                    > The `prior_identifier` must be unique since it is used to identify the prior.

                    > When generating the results, if priors are used the result includes
                    prior=`prior_identifier` for identification.

                    > We suggest the `prior_identifier` to be the name of the objective(s)
                    over which the prior is defined + some unique identifier
                    (eg: value1_good_value2_bad).

    Returns:
        The result of the _run function as a pandas DataFrame.
    """
    if isinstance(benchmark, FunctionalBenchmark):
        benchmark = benchmark.desc

    if priors:
        priors = dict_to_configpriors(priors)

    optimizer_hyperparameters = optimizer_hyperparameters or {}

    problem = Problem.problem(
        optimizer=optimizer,
        optimizer_hyperparameters=optimizer_hyperparameters,
        benchmark=benchmark,
        objectives=objectives,
        fidelities=fidelities,
        budget=budget,
        continuations=continuations,
        priors=priors,
    )

    history = _run(
        run_name=run_name,
        problem=problem,
        seed=seed,
    )
    _df = pd.DataFrame([res._to_dict() for res in history])
    fidelities = problem.get_fidelities()
    match fidelities:
        case None:
            _fidelities = None
        case str():
            _fidelities = [fidelities] * len(_df)
        case list():
            _fidelities = fidelities * len(_df)
        case _:
            raise ValueError(f"Unsupported fidelities type: {type(fidelities)}")

    costs = problem.get_costs()
    match costs:
        case None:
            _costs = None
        case str():
            _costs = [costs] * len(_df)
        case list():
            _costs = costs * len(_df)
        case _:
            raise ValueError(f"Unsupported costs type: {type(costs)}")

    objs = problem.get_objectives()
    match objs:
        case None:
            _objectives = None
        case str():
            _objectives = [objs] * len(_df)
        case list():
            _objectives = objs * len(_df)
        case _:
            raise ValueError(f"Unsupported objectives type: {type(objs)}")

    _df = _df.assign(
        seed=seed,
        optimizer=problem.optimizer.name,
    )
    if len(problem.optimizer_hyperparameters) > 0:
        _df["optimizer_hps"] = ",".join(
            f"{k}={v}" for k, v in problem.optimizer_hyperparameters.items()
        )
    else:
        _df["optimizer_hps"] = "default"
    return _df.assign(
        benchmark=problem.benchmark.name,
        objectives=_objectives,
        fidelities=fidelities,
        costs=costs,
    )


from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any

from hpoglue.result import Result

if TYPE_CHECKING:
    from hpoglue.config import Config


@dataclass(kw_only=True)
class Query:
    """A query to a benchmark."""

    config: Config
    """ The config to evaluate """

    fidelity: tuple[str, int | float] | Mapping[str, int | float] | None = None
    """What fidelity to evaluate at."""

    optimizer_info: Any | None = None
    """Any optimizer specific info required across ask and tell"""

    request_trajectory: bool | tuple[int, int] | tuple[float | float] = False
    """Whether the optimizer requires a trajectory curve for multi-fidelity optimization.

    If a specific range is requested, then a tuple can be provided.
    """

    query_id: str = field(init=False)
    """The id of the query.

    This includes information about the config id and fidelity.
    """

    def __post_init__(self) -> None:
        match self.fidelity:
            case None:
                self.query_id = self.config.config_id
            case (name, value):
                self.query_id = f"{self.config.config_id}-{name}={value}"
            case Mapping():
                self.query_id = (
                    f"{self.config.config_id}-"
                    f"{'-'.join(f'{k}={v}' for k, v in self.fidelity.items())}"
                )
            case _:
                raise NotImplementedError("Unexpected fidelity type")

        if self.request_trajectory:
            match self.fidelity:
                case None:
                    raise ValueError("Learning curve requested but no fidelity provided")
                case tuple():
                    pass
                case Mapping():
                    raise ValueError(
                        "Learning curve requested but more than a single fidelity provided"
                    )

    @property
    def config_id(self) -> str:
        """The id of the config."""
        return self.config.config_id

    def with_fidelity(
        self,
        fidelity: tuple[str, int | float] | Mapping[str, int | float] | None,
    ) -> Query:
        """Create a new query with a different fidelity."""
        return replace(self, fidelity=fidelity)

    def make_result(self, results: dict[str, Any]) -> Result:
        """Create a result from the query."""
        return Result(
            query=self,
            fidelity=self.fidelity,
            values=results,
        )

    def _to_dict(self) -> dict[str, Any]:
        return {
            "config_id": self.config_id,
            "config": self.config.values,
            "fidelity": self.fidelity,
        }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, TypeAlias


@dataclass(frozen=True)
class TrialBudget:
    """A budget for the number of trials to run."""

    name: ClassVar[str] = "trial_budget"

    total: int
    """Total amount of budget allowed for the optimizer for this problem.

    How this is interpreted is depending on fidelity type.

    If the problem **does not** include a fidelity, then this is assumed
    to be a black-box problem, and each fully complete trial counts as
    1 towards the budget.

    If the problem **does** include a **single** fidelity, then the fidelity
    at which the trial was evaluated is taken as a fraction of the full fidelity
    and added to the used budget. For example, 40 epochs of a fidelity that
    maxes out at 100 epochs would count as 0.4 towards the budget.

    If the problem **does** include **many** fidelities, then the fraction as calculated
    for a single fidelity is applied to all fidelities, and then summed, normalized by
    the total number of fidelities. For example, 40 epochs of a fidelity that maxes out
    at 100 epochs and data percentage of 0.6 of a fidelity that maxes out at 1.0 would
    equate to (0.4 + 0.6) / 2 = 0.5 towards the budget.
    """

    minimum_fidelity_normalized_value: float = field(default=0.01, repr=False)

    @property
    def path_str(self) -> str:
        """Return a string representation of the budget."""
        clsname = self.__class__.__name__
        return f"{clsname}={self.total}"

    def to_dict(self) -> dict[str, int]:
        """Convert the budget to a dictionary."""
        return {"total": self.total}

    @classmethod
    def from_dict(cls, data: dict[str, int]) -> TrialBudget:
        """Convert a dictionary to a budget."""
        return cls(total=data["total"])


@dataclass(frozen=True)
class CostBudget:
    """A budget for the cost of the trials to run."""

    name: ClassVar[str] = "cost_budget"

    total: int | float

    def __post_init__(self):
        raise NotImplementedError("Cost budgets not yet supported")

    def to_dict(self) -> dict[str, int | float]:
        """Convert the budget to a dictionary."""
        return {"total": self.total}

    @classmethod
    def from_dict(cls, data: dict[str, int | float]) -> CostBudget:
        """Convert a dictionary to a budget."""
        return cls(total=data["total"])

    @property
    def path_str(self) -> str:
        """Return a string representation of the budget."""
        clsname = self.__class__.__name__
        return f"{clsname}={self.total}"


BudgetType: TypeAlias = TrialBudget | CostBudget



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

import numpy as np


class OutOfBoundsError(ValueError):
    """Raised when a value is outside of the bounds of a measure."""


@dataclass(kw_only=True, frozen=True)
class Measure:
    """A measure for a becnhmark.

    It's main use is to convert a raw value into a value that can always be
    minimized.
    """

    minimize: bool
    """Whether or not to smaller is better."""

    kind: Measure.Kind
    """What kind of measure this is."""

    bounds: tuple[float, float] = field(default_factory=lambda: (-np.inf, np.inf))
    """The bounds of the measure."""

    @classmethod
    def metric(
        cls,
        bounds: tuple[float, float],
        *,
        minimize: bool,
    ) -> Measure:
        """Create a metric measure.

        Args:
            minimize: Whether or not to minimize the value.
            bounds: The bounds of the metric.

        Returns:
            The metric measure.
        """
        return cls(minimize=minimize, kind=cls.Kind.METRIC, bounds=bounds)

    @classmethod
    def cost(cls, bounds: tuple[float, float], *, minimize: bool = True) -> Measure:
        """Create a cost measure.

        Args:
            bounds: The bounds of the cost.
            minimize: Whether or not to minimize the value.

        Returns:
            The cost measure.
        """
        return cls(minimize=minimize, kind=cls.Kind.COST, bounds=bounds)

    @classmethod
    def test_metric(
        cls,
        bounds: tuple[float, float],
        *,
        minimize: bool,
    ) -> Measure:
        """Create a test metric measure.

        Args:
            minimize: Whether or not to minimize the value.
            bounds: The bounds of the metric.

        Returns:
            The metric measure.
        """
        return cls(minimize=minimize, kind=cls.Kind.TEST_METRIC, bounds=bounds)

    class Kind(str, Enum):
        """Enumeration of measure kinds."""
        METRIC = "metric"
        COST = "cost"
        TEST_METRIC = "test_metric"

        def __str__(self) -> str:
            return self.value

    def __post_init__(self) -> None:
        if self.bounds[0] >= self.bounds[1]:
            raise ValueError(
                f"bounds[0] must be less than bounds[1], got {self.bounds}",
            )

    @property
    def optimum(self) -> float:
        """Get the optimum value for this metric.

        Returns:
            The optimum value.
        """
        match self.minimize:
            case True:
                return self.bounds[0]
            case False:
                return self.bounds[1]

    def as_minimize(self, value: float) -> float:
        """Convert a raw value into a value that should be minimized."""
        match self.minimize:
            case True:
                return float(value)
            case False:
                return -float(value)

    def as_maximize(self, value: float) -> float:
        """Convert a raw value into a value that should be maximized."""
        match self.minimize:
            case True:
                return -float(value)
            case False:
                return float(value)



from __future__ import annotations

from collections.abc import Iterator, Sequence, Sized
from dataclasses import dataclass
from typing import Generic, Protocol, TypeVar, runtime_checkable
from typing_extensions import Self

import numpy as np

T = TypeVar("T", bound=int | float)


@runtime_checkable
class Fidelity(Protocol[T]):
    """A protocol to represent a Fidelity type, which includes a minimum and maximum value of a
    specific type, along with additional metadata such as the type of the values and whether
    continuation is supported.
    """

    kind: type[T]
    min: T
    max: T
    supports_continuation: bool

    @classmethod
    def frm(
        cls,
        values: tuple[T, T] | tuple[T, T, T] | Sequence[T] | range,
        *,
        supports_continuation: bool = False,
    ) -> Fidelity:
        """Create a Fidelity instance from a tuple, sequence, or range object.

        Args:
            values: A tuple of two or three values, a sequence of values, or a range object.

                - If a tuple of two values, the values are the minimum and maximum values,
                    considered on a conitnuous range.
                - If a tuple of three values, the values are the minimum, maximum, and step size,
                  giving a discrete number of fidelity values.
                - If a sequence of values, the values are the fidelity values, which is just
                    a list of values allowed
                - If a range object, the values are the start, stop, and step values.

            supports_continuation: A flag indicating if continuation is supported.
                Defaults to False.

        Returns:
            Fidelity: An instance of Fidelity with the specified values.
        """
        match values:
            case range():
                return RangeFidelity.from_range(values, supports_continuation=supports_continuation)
            case tuple() if len(values) == 2:  # noqa: PLR2004
                return ContinuousFidelity.from_tuple(
                    values, supports_continuation=supports_continuation
                )
            case tuple() if len(values) == 3:  # noqa: PLR2004
                return RangeFidelity.from_tuple(values, supports_continuation=supports_continuation)
            case Sequence():
                return ListFidelity.from_seq(values, supports_continuation=supports_continuation)
            case _:
                raise ValueError(f"Unsupported values type: {type(values)}")


@dataclass(kw_only=True, frozen=True)
class ListFidelity(Fidelity, Sized, Generic[T]):
    """A class to represent a List Fidelity type, which includes a sorted list of values of a
    specific type, along with additional metadata such as the minimum and maximum values,
    and whether the list supports continuation.

    Attributes:
        kind: The type of the elements in the list.

        values: A sorted tuple of values.

        min: The minimum value in the list.

        max: The maximum value in the list.

        supports_continuation: A boolean flag indicating if the list supports continuation.
    """

    kind: type[T]
    values: tuple[T, ...]
    min: T
    max: T
    supports_continuation: bool

    def __len__(self) -> int:
        return len(self.values)

    @classmethod
    def from_seq(
        cls,
        values: Sequence[T],
        *,
        supports_continuation: bool = False,
    ) -> ListFidelity[T]:
        """Create a ListFidelity instance from a sequence of values.

        Args:
            values: The sequence of values to create the ListFidelity from.

            supports_continuation: Indicates if continuation is supported.
                Defaults to False.

        Returns:
            ListFidelity: An instance of ListFidelity containing the sorted values.
        """
        vs = sorted(values)
        return cls(
            kind=type(vs[0]),
            values=tuple(vs),
            supports_continuation=supports_continuation,
            min=vs[0],
            max=vs[-1],
        )

    def __iter__(self) -> Iterator[T]:
        return iter(self.values)


@dataclass(kw_only=True, frozen=True)
class RangeFidelity(Fidelity, Sized, Generic[T]):
    """A class to represent a Range Fidelity type that iterates over a range of values with a
        specified step size.

    Attributes:
        kind: The type of the range values (int or float).

        min: The minimum value of the range.

        max: The maximum value of the range.

        stepsize: The step size for iterating over the range.

        supports_continuation: A boolean flag indicating if continuation is supported.
    """

    kind: type[T]
    min: T
    max: T
    stepsize: T
    supports_continuation: bool

    def __post_init__(self):
        if self.min >= self.max:
            raise ValueError(f"min must be less than max, got {self.min} and {self.max}")

        if self.stepsize <= 0:
            raise ValueError(f"stepsize must be greater than 0, got {self.stepsize}")

        # Ensure bounds quantize correctly into `n_values` stepsize chunks
        n_values = int((self.max - self.min) / self.stepsize) + 1
        if not np.isclose(self.min + (n_values * self.stepsize) - 1, self.max):
            raise ValueError(
                f"stepsize {self.stepsize} does not divide range [{self.min}, {self.max}]"
            )

    def __iter__(self) -> Iterator[T]:
        current = self.min
        yield self.min
        while current < self.max:
            current += self.stepsize
            yield max(current, self.max)  # type: ignore

    @property
    def n_values(self) -> int:
        """The number of values in the range."""
        return int((self.max - self.min) / self.stepsize) + 1

    def __len__(self) -> int:
        return self.n_values

    @classmethod
    def from_tuple(
        cls,
        values: tuple[T, T, T],
        *,
        supports_continuation: bool = False,
    ) -> RangeFidelity[T]:
        """Create a RangeFidelity instance from a tuple of values.

        Args:
            values: A tuple containing three values of type T (int or float).

            supports_continuation: A flag indicating if continuation is supported.
                Defaults to False.

        Returns:
            RangeFidelity: An instance of RangeFidelity with the specified values.

        Raises:
            ValueError: If the values are not all of type int or float,
                or if the values are not of the same type.
        """
        _type = type(values[0])
        if _type not in (int, float):
            raise ValueError(f"all values must be of type int or float, got {_type}")

        if len(values) != 3:  # noqa: PLR2004
            raise ValueError(f"expected 3 values, got {len(values)}")

        if not all(isinstance(v, _type) for v in values):
            raise ValueError(f"all values must be of type {_type}, got {values}")

        return cls(
            kind=_type,
            min=values[0],
            max=values[1],
            stepsize=values[2],
            supports_continuation=supports_continuation,
        )

    @classmethod
    def from_range(cls, r: range, *, supports_continuation: bool = False) -> Self:
        """Create a RangeFidelity instance from a range object.

        Args:
            r: A range object.
            supports_continuation: A flag indicating if continuation is supported.

        Returns:
            RangeFidelity: An instance of RangeFidelity with the specified values.
        """
        return cls(
            kind=int,
            min=r.start,
            max=r.stop,
            stepsize=r.step,
            supports_continuation=supports_continuation,
        )


@dataclass(kw_only=True, frozen=True)
class ContinuousFidelity(Fidelity, Generic[T]):
    """A class to represent a continuous fidelity range with a minimum and maximum value.

    Attributes:
        kind: The type of the fidelity values (int or float).

        min: The minimum value of the fidelity range.

        max: The maximum value of the fidelity range.

        supports_continuation: A boolean flag indicating if continuation is supported.
    """

    kind: type[T]
    min: T
    max: T
    supports_continuation: bool

    def __post_init__(self):
        if self.min >= self.max:
            raise ValueError(f"min must be less than max, got {self.min} and {self.max}")

    @classmethod
    def from_tuple(
        cls,
        values: tuple[T, T],
        *,
        supports_continuation: bool = False,
    ) -> ContinuousFidelity[T]:
        """Create a ContinuousFidelity instance from a tuple of values.

        Args:
            values: A tuple containing two values of type int or float.

            supports_continuation: A flag indicating if continuation is supported.
                Defaults to False.

        Returns:
            ContinuousFidelity: An instance of ContinuousFidelity with the specified values.

        Raises:
            ValueError: If the values are not of type int or float,
                or if the values are not of the same type.
        """
        _type = type(values[0])
        if _type not in (int, float):
            raise ValueError(f"all values must be of type int or float, got {_type}")

        if len(values) != 2:  # noqa: PLR2004
            raise ValueError(f"expected 2 values, got {len(values)}")

        if not all(isinstance(v, _type) for v in values):
            raise ValueError(f"all values must be of type {_type}, got {values}")

        return cls(
            kind=_type,
            min=values[0],
            max=values[1],
            supports_continuation=supports_continuation,
        )
